@@ -4,8 +4,6 @@
 // Package osdiag provides loggers for OS-specific diagnostic information.
 package osdiag
 
-import "tailscale.com/types/logger"
-
 // LogSupportInfoReason is an enumeration indicating the reason for logging
 // support info.
 type LogSupportInfoReason int
@@ -15,9 +13,8 @@ const (
 	LogSupportInfoReasonBugReport                                 // a bugreport is in the process of being gathered.
 )
 
-// LogSupportInfo obtains OS-specific diagnostic information useful for
-// troubleshooting and support, and writes it to logf. The reason argument is
-// useful for governing the verbosity of this function's output.
-func LogSupportInfo(logf logger.Logf, reason LogSupportInfoReason) {
-	logSupportInfo(logf, reason)
+// SupportInfo obtains OS-specific diagnostic information for troubleshooting
+// and support. The reason governs the verbosity of the output.
+func SupportInfo(reason LogSupportInfoReason) map[string]any {
+	return supportInfo(reason)
 }
